import {
    calcFrameCheckSum,
    convertByte2Len
} from './util';

import {
    MEOS_FRAME_HEAD,
    MEOS_FRAME_END
} from './setting';

/**
 * 负责返回协议的解码
 */
const Decodebytes = {
    //缓存已接收值
    responseCache: [],
    /**
     * 剔除头部之前的脏数据
     * @param  {Array} totalData  缓存数据 + 本次数据
     */
    getCleanFrame_(totalData) {
        return totalData[0] === MEOS_FRAME_HEAD ? totalData : totalData.slice(totalData.indexOf(MEOS_FRAME_HEAD));
    },

    /**
     * 解析 Cody 协议帧
     * @param  {Array} currentData 当前协议包（片段）
     * @param  {Boolean} splitAllFrame 是否分离出协议包的所有帧
     * @exmaple
     *
     * @return {Array|NULL}
     */
    splitFrame(currentData, splitAllFrame = true) {
        let totalData = this.responseCache.concat(currentData);
        // 无帧头
        if (!totalData.includes(MEOS_FRAME_HEAD)) {
            return null;
        }
        let tempFrame_ = this.getCleanFrame_(totalData);
        // 长度过小
        if(tempFrame_.length < 7) {
            this.responseCache = tempFrame_;
            return undefined;
        }
        // 帧数据的标记长度 2bytes
        let frameDataLen = convertByte2Len(tempFrame_[2], tempFrame_[3]);
        let localFrameHeaderCheck = calcFrameCheckSum([tempFrame_[0], tempFrame_[2], tempFrame_[3]]);
        // 头部校验不通过 - 无效数据
        if(localFrameHeaderCheck !== tempFrame_[1]) {
            this.responseCache = [];
            return null;
        }
        // 帧总长度
        let totalLen = tempFrame_.length;
        // 无尾部出现 - 等待下一次数据
        if (!tempFrame_.includes(MEOS_FRAME_END)) {
            // 错误数据
            // [frame_head, head_checksum, len1, len2, checksum].length => 5
            if (totalLen - 5 > frameDataLen) {
                this.responseCache = [];
                return null;
            }
            this.responseCache = tempFrame_;
            return undefined;
        }
        //有头有尾
        else {
            // 可能尾部过早出现 - 等待下一次数据
            if (totalLen - 5 <= frameDataLen) {
                this.responseCache = tempFrame_;
                return undefined;
            }
            // 清空
            this.responseCache = [];
            // 数据包含多个帧时，需做切割
            // [242,0xf9,7,0,  0x02,0xf9,97,0,9,12,2,212,65,136,245,  242,0xf9,7,0,  0x02,97,0,9,51,51,212,65,136,245,  242,0xf9,7,0,  0x02,97,0,9,66,96,212,65,40]
            let resultFrame = [];
            let parseAllFrame = true;
            // 每一帧的起始索引
            let eachFrameIndex = 0;
            // 一个 frame 的数据部分长度
            let eachFrameDataLen = 0;
            do {
                eachFrameDataLen = convertByte2Len(tempFrame_[eachFrameIndex + 2], tempFrame_[eachFrameIndex + 3]);
                // 一个完整 frame 长度
                let frameLen = eachFrameDataLen + 6;
                // 注：考虑实际需求和性能平衡，限定值解析一帧数据
                if (parseAllFrame) {
                    parseAllFrame = splitAllFrame;
                    let checkSum = tempFrame_[eachFrameIndex + eachFrameDataLen + 4];
                    let frameEnd = tempFrame_[eachFrameIndex + eachFrameDataLen + 5];
                    let localCheckSum = calcFrameCheckSum(tempFrame_.slice(eachFrameIndex + 4, eachFrameIndex + eachFrameDataLen + 4));
                    // 正确数据
                    if (frameEnd === MEOS_FRAME_END && checkSum === localCheckSum) {
                        resultFrame.push(tempFrame_.slice(eachFrameIndex, eachFrameIndex + frameLen));
                    }
                }
                //下一帧
                eachFrameIndex += frameLen;
                eachFrameDataLen = convertByte2Len(tempFrame_[eachFrameIndex + 2], tempFrame_[eachFrameIndex + 3]);
            }
            while (eachFrameDataLen && totalLen - eachFrameIndex >= eachFrameDataLen + 6);

            // 剩余部分不足下一帧 - 缓存起来下次使用
            if (totalLen - eachFrameIndex < eachFrameDataLen + 6) {
                this.responseCache = tempFrame_.slice(eachFrameIndex);
            }
            return resultFrame.length ? resultFrame : null;
        }
    }
}

export default Decodebytes;