import {
    bytesToString,
    ascii2Str,
    intArrayToHexArray,
    convertByte2Len,
    calculateResponseValue,
    bytesToInt
} from './util';

import {
    MAC_TYPE,
    HANDSHAKE_DTR,
    TIPS,
    WORKING_MODE,
    WORKING_MODE_CMD,
    CORRECTION_MODE,
    DO_CORRECT
} from './setting';

import utf8 from 'utf8';

/**
 * 负责返回协议的解析
 */
const Parse = {
    logCache: [],
    /**
     * 从协议中提取变量名、数值
     */
    getNameAndValue(buff) {
        let terminal = buff.slice(5).indexOf(0) + 5; // 排除头部固定的5个位以后，首次出现 0x00 时
        let nameBytes = buff.slice(5, terminal);
        let dataType = buff[terminal + 1];
        let value = null;
        switch (dataType) {
            case 1:
                // 1byte
                value = buff[terminal + 2];
                break;
            case 11:
                // 1byte boolean
                value = !!buff[terminal + 2];
                break;
            case 3:
            case 4:
                // 2byte short
                value = calculateResponseValue(buff.slice(terminal + 2, terminal + 4).reverse());
                break;
            case 12:
                // 4byte string
                value = bytesToString(buff.slice(terminal + 2, buff.length - 2));
                break;
            case 5:
                // 4byte int
                value = bytesToInt(buff.slice(terminal + 2, terminal + 6).reverse());
                break;
            case 6:
            case 9:
                // long型或者float型的4byte处理
                value = calculateResponseValue(buff.slice(terminal + 2, terminal + 6).reverse());
                break;
            default:
                break;
        }
        // console.log(buff, '=>', ascii2Str(nameBytes), value);
        let name = ascii2Str(nameBytes);
        name = utf8.decode(name);
        return {
            name: name,
            value
        };
    },

    /**
     * 从协议中解析传感器的值
     * @example
     * [243,251,8,0,   4, 2,  44,1,  0,0,0,0,   51,244]
     * @return {[type]} [description]
     */
    getSensorVal(buff) {
        let id = convertByte2Len(buff[6], buff[7]);
        let action = buff[5];
        let result = {id, action};
        switch (action) {
            case 3:
                // long型或者float型的4byte处理
                result.value = calculateResponseValue(buff.slice(8, 12).reverse());
                break;
            case 4:
            case 5:
                break;
            default:
                break;
        }
        return result;
    },

    /**
     * 从协议中解析广播信息
     * @example
     * [243,246,3,0,  0x03, 97,0  100,244]
     */
    getBroadcastMsg(buff) {
        let len = convertByte2Len(buff[2], buff[3]);
        let msg = bytesToString(buff.slice(5, len+3));
        msg = utf8.decode(msg);
        return msg;
    },

    /**
     * [getMacAddr description]
     * @param  {Array} buff [description]
     * @example
     * [243, 250, 7, 0, 5, 2, 48, 174, 164, 39, 222, 142, 244]
     * @return {[type]}      [description]
     */
    getMacAddr(buff) {
        let macSlice = buff.slice(6, 12);
        let mac = intArrayToHexArray(macSlice).map((val) => val.toUpperCase()).join(' ');
        return {mac, type: MAC_TYPE[buff[5]]};
    },
    /**
     * [getVersion description]
     * @param  {Array} buff [description]
     * @example
     * [243, 250, 7, 0, 0x06, xx xx xx xx xx, 142, 244]
     * @return {[type]}      [description]
     */
    getVersion(buff) {
        let len = convertByte2Len(buff[2], buff[3]);
        let version = bytesToString(buff.slice(5, len+4));
        version = utf8.decode(version);
        return {version, name: TIPS.codey};
    },

    /**
     * [getHandshakeMsg description]
     * @param  {Array} buff [description]
     * @example
     * dtr is ready: [243, 245, 2, 0, 0x07, 0x02, 9, 244]
     * @return {[type]}      [description]
     */
    getHandshakeMsg(buff) {
        if(buff[5] === HANDSHAKE_DTR.RESPONSE_DTR) { // dtr is ready
            return {
                status: 20,
                tips: TIPS.dtr_ready
            }
        }
    },

    /**
     * [getReadyResult description]
     * @param  {Array} buff [description]
     * @example
     * @return {[type]}      [description]
     */
    getReadyResult(buff) {
        return {
            type: buff[5]
        }
    },

    getCommChannel(buff) {
        let len_ = buff[7]; // 长度位
        let channel = bytesToString(buff.slice(8, 8+len_));
        channel = utf8.decode(channel);
        return {
            channel
        }
    },

    getWorkingMode(buff) {
        let mode = WORKING_MODE[buff[6]];
        return {
            mode
        }
    },

    getSensorCorrection(buff) {
        let sub_id = buff[5];
        if (sub_id === CORRECTION_MODE.SUB_ID) { // 进入、退出
            let mode = buff[6];
            let result = buff[7];
            return {
                mode,
                result
            }
        } else if (sub_id === DO_CORRECT.SUB_ID) { // 开始校正
            let sensorId = buff[6];
            let result = buff[7];
            return {
                sensorId,
                sensorName: DO_CORRECT.SENSOR[sensorId],
                result
            }
        }
    },

    getlogInfo (logBuff) {
        let logs = this.logCache.concat(logBuff);
        let log13 = logs.lastIndexOf(13);
        let log10 = logs.lastIndexOf(10);
        if(log13 + 1 === log10) {
            this.logCache = logs.slice(log10+1);
            return logs.slice(0, log10+1);
        }else {
            return [];
        }
    }
}

export default Parse;