import {
    stringToAsciiCode,
    arrayFromArrayBuffer,
    redoAsync
} from './util';
import Command from './command';
import Encodebytes from './encodebytes';
import {genEvent} from './emitter';
import {
    EVENT
} from './setting';

import utf8 from 'utf8';

class Uploader {
    /**
     * Create a uploader.
     */
    constructor(emiter) {
        this.currentContent_ = null;
        this.currentPath_ = null;
        // 每次最大的传输字符长度
        this.file_chunk_length_ = 80;
        this.totalSize = 0;
        this.uploadedSize = 0;
        this.emiter = emiter;
    }

    /**
     * setter 对外接口，外部模块调用此方法，设置将要上传给 meos 的内容
     * @param  {String} ctt 上传内容
     */
    set content(ctt = '') {
        this.currentContent_ = ctt;
    }

    get content() {
        return this.currentContent_;
    }

    /**
     * 设置上传路径
     * @param  {String} url [description]
     * @return {[type]}     [description]
     */
    set path(url) {
        if(typeof url == 'string') {
            this.currentPath_ = url;
        }
    }

    get path() {
        return this.currentPath_;
    }

    /**
     * 设置文件块长度，以定制传输速度
     * TOFIX: 长度需在合法范围值中
     * @param  {Number} len default is 80 bytes
     */
    set blockLength(len = 80) {
        this.file_chunk_length_ = len;
    }

    get blockLength() {
        return this.file_chunk_length_;
    }

    /**
     * 上传内容
     * 前提是要设置上传内容
     * 文件名最大长度 256，暂时未做校验
     * 代码文件路径为：'/flash/main.py'
     * 音乐文件路径为：'/music/cat.wav'
     */
    uploadContent() {
        if(!this.path) {
            console.warn('please set the path before uploading.')
            return;
        }
        if(!this.content) {
            console.warn('please set the content before uploading.')
            return;
        }
        let contentBytes = [];
        // 传声音
        if(/\.wav$/.test(this.path)) {
            if(this.content instanceof Array) {
                contentBytes = this.content;
            }else if(this.content instanceof ArrayBuffer){ // arrayBuffer
                contentBytes = arrayFromArrayBuffer(this.content);
            }else {
                console.warn('content is illegal');
            }
        }
        // 传 code
        else if(/\.py$/.test(this.path) && typeof this.content === 'string') {
            // 做中文处理：使用 utf8 库将 js 字符编码转成 utf8，js 字符串默认编码是 ucs
            let content = utf8.encode(this.content)
                contentBytes = stringToAsciiCode(content);
        }
        this.uploadFile(this.path, contentBytes);
    }

    /**
     * 上传
     * @interface
     * @param  {String} path     文件路径
     * @param  {ArrayBuffer} contentBytes [description]
     * @return {[type]}              [description]
     */
    async uploadFile(path, contentBytes) {
        let allBytes = Encodebytes.composer(path, contentBytes, this.blockLength);
        // 上传前置空
        this.totalSize = 0;
        this.uploadedSize = 0;
        // 计算总大小
        allBytes.forEach((val) => {
            this.totalSize += val.length;
        });
        // 分块发送
        for (let i = 0, bytes; bytes = allBytes[i]; i++) {
            let result;
            if(i == 0) {
                // 外部需求：通过尝试发送确认重启是否已 ready，2次不成功则表示固件故障
                result = await redoAsync(() => Command.upload(bytes), 2);
            }else {
                result = await Command.upload(bytes);
            }
            //向外汇报进度
            this.progressHandler_(result.err, bytes);
            if (result.err) {
                return;
            }
        }
    }

    /**
     * TODO: 删除已上传文件
     * @interface
     * @return {Promise} [description]
     */
    async deleteFile(path) {
        let bytes = Encodebytes.deleteFileBytes(path);
        let result = await Command.upload(arrayBytes);
        return Promise.resolve(result.err?'success':'fail');
    }

    /**
     * 查询已上传文件
     * @interface
     * @return {[type]} [description]
     */
    async queryFile() {
        // TODO
    }

    /**
     * 外部模块想获取发送进度数据，需用此接口传递监听函数
     * @param {Function} handler [description]
     * @param {Boolean} isOnce 是否只监听依次
     */
    onUploadProgress(handler, isOnce) {
        return genEvent(EVENT.UPLOAD_PROGRESS, this.emiter)(handler, isOnce);
    }

    /**
     * 上传进度处理函数
     * @private
     * @param  {Null|Object} error [description]
     * @param  {Array} bytes [description]
     */
    progressHandler_(error, bytes) {
        let data = {
            total: this.totalSize,
            bytes: bytes,
            uploaded: 0
        };
        this.uploadedSize += bytes.length;
        if (this.totalSize !== 0) {
            data.uploaded = Math.min(Math.floor(this.uploadedSize), this.totalSize);
        }
        //对内部进度数据分发
        this.emiter.emit(EVENT.UPLOAD_PROGRESS, error, data);
    }

}

export default Uploader;